"""Base functions for universal machine-learned interatomic potentials."""
from __future__ import annotations

import warnings
from functools import lru_cache
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Literal

    from ase.calculator.calculator import Calculator


@lru_cache
def _pick_calculator(
    calculator: Literal["mace", "m3gnet", "chgnet"], **kwargs
) -> Calculator:
    """
    Adapted from `matcalc.util.get_universal_calculator`.

    Parameters
    ----------
    calculator
        Name of the calculator to use
    **kwargs
        Custom kwargs for the underlying calculator. Set a value to
        `quacc.Remove` to remove a pre-existing key entirely. For a list of available
        keys, refer to the `mace.calculators.mace_mp`, `chgnet.model.dynamics.CHGNetCalculator`,
        or `matgl.ext.ase.M3GNetCalculator` calculators.

    Returns
    -------
    Calculator
        The chosen calculator
    """
    import torch

    if not torch.cuda.is_available():
        warnings.warn("CUDA is not available to PyTorch. Calculations will be slow.")

    if calculator.lower().startswith("m3gnet"):
        import matgl
        from matgl import __version__
        from matgl.ext.ase import M3GNetCalculator

        model = matgl.load_model("M3GNet-MP-2021.2.8-DIRECT-PES")
        kwargs.setdefault("stress_weight", 1.0 / 160.21766208)
        calc = M3GNetCalculator(potential=model, **kwargs)

    elif calculator.lower() == "chgnet":
        from chgnet import __version__
        from chgnet.model.dynamics import CHGNetCalculator

        calc = CHGNetCalculator(**kwargs)

    elif calculator.lower() == "mace":
        from mace import __version__
        from mace.calculators import mace_mp

        calc = mace_mp(**kwargs)

    else:
        raise ValueError(f"Unrecognized {calculator=}.")

    calc.parameters["version"] = __version__

    return calc



"""
Utilities for working with lists.
"""
from __future__ import annotations


def merge_list_params(
    *lists: list[str], removal_prefix: str | None = "#", case_insensitive: bool = True
) -> list[str]:
    """
    Merge list string parameters, taking list2 as the priority.
    Also removes any entries that start with `removal_prefix` from the final list.

    Parameters
    ----------
    *lists
        Lists to merge, with the latter taking priority.
    removal_prefix
        Prefix to use to remove an entry from the final list.
    case_insensitive
        Whether to perform case-insensitive comparisons

    Returns
    -------
    list
        Merged list
    """

    merged_list = []
    lists = [list_ for list_ in lists if list_]
    for list_ in lists:
        for item in list_:
            if case_insensitive:
                item = item.lower()
            if item not in merged_list:
                merged_list.append(item)

    if removal_prefix:
        items_to_remove1 = [
            item[1:] for item in merged_list if item.startswith(removal_prefix)
        ]
        items_to_remove2 = [
            item for item in merged_list if item.startswith(removal_prefix)
        ]
        items_to_remove = items_to_remove1 + items_to_remove2
        for item in items_to_remove:
            if item in merged_list:
                merged_list.remove(item)

    merged_list.sort()
    return merged_list



"""
Parameter-related utilities for the Q-Chem calculator.
"""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ase import Atoms
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.io.qchem.inputs import QCInput
from pymatgen.io.qchem.sets import QChemDictSet
from pymatgen.io.qchem.utils import lower_and_check_unique

from quacc.utils.dicts import recursive_dict_merge, sort_dict

if TYPE_CHECKING:
    from typing import Any

    from quacc.calculators.qchem.qchem import QChem

logger = logging.getLogger(__name__)


def make_qc_input(qchem: QChem, atoms: Atoms) -> QCInput:
    """
    Make a QCInput object. It will, by default, create a QCInput from
    the QChem calculator kwargs. If `qchem.qchem_dict_set_params` is
    specified, it will create a QCInput from a QChemDictSet, merging
    the two QCInput objects and taking the latter as higher priority.

    Parameters
    ----------
    qchem
        The QChem object.

    Returns
    -------
    QCInput
        The QCInput object.
    """

    atoms.charge = qchem.charge
    atoms.spin_multiplicity = qchem.spin_multiplicity
    molecule = AseAtomsAdaptor().get_molecule(atoms)

    if qchem.qchem_dict_set_params:
        # Get minimal parameters needed to instantiate a QChemDictSet
        if "molecule" in qchem.qchem_dict_set_params:
            msg = "Do not specify `molecule` in `qchem_dict_set_params`"
            raise NotImplementedError(msg)
        if "job_type" not in qchem.qchem_dict_set_params and qchem.rem.get("job_type"):
            qchem.qchem_dict_set_params["job_type"] = qchem.rem["job_type"]
        if "basis_set" not in qchem.qchem_dict_set_params and qchem.rem.get("basis"):
            qchem.qchem_dict_set_params["basis_set"] = qchem.rem["basis"]
        if "scf_algorithm" not in qchem.qchem_dict_set_params and qchem.rem.get(
            "scf_algorithm"
        ):
            qchem.qchem_dict_set_params["scf_algorithm"] = qchem.rem["scf_algorithm"]
        if "qchem_version" not in qchem.qchem_dict_set_params:
            qchem.qchem_dict_set_params["qchem_version"] = 6

        # Make QChemDictSet
        qc_dict_set = QChemDictSet(molecule, **qchem.qchem_dict_set_params)
        for prop in [
            "rem",
            "opt",
            "pcm",
            "solvent",
            "smx",
            "scan",
            "van_der_waals",
            "plots",
            "nbo",
            "geom_opt",
            "svp",
            "pcm_nonels",
        ]:
            prop1 = getattr(qchem, prop)
            if prop2 := getattr(qc_dict_set, prop):
                setattr(qchem, prop, recursive_dict_merge(prop2, prop1))
        for prop in ["vdw_mode", "cdft", "almo_coupling"]:
            prop2 = getattr(qc_dict_set, prop)
            if prop2 and not prop1:
                setattr(qchem, prop, prop2)

    qchem.rem = sort_dict(
        get_rem_swaps(qchem.rem, restart=qchem.prev_orbital_coeffs is not None)
    )

    return QCInput(
        molecule,
        qchem.rem,
        opt=qchem.opt,
        pcm=qchem.pcm,
        solvent=qchem.solvent,
        smx=qchem.smx,
        scan=qchem.scan,
        van_der_waals=qchem.van_der_waals,
        vdw_mode=qchem.vdw_mode,
        plots=qchem.plots,
        nbo=qchem.nbo,
        geom_opt=qchem.geom_opt,
        cdft=qchem.cdft,
        almo_coupling=qchem.almo_coupling,
        svp=qchem.svp,
        pcm_nonels=qchem.pcm_nonels,
    )


def cleanup_attrs(qchem: QChem) -> None:
    """
    Clean up self attribute parameters in place.

    Parameters
    ----------
    qchem
        The QChem object

    Returns
    -------
    None
    """
    for attr in [
        "rem",
        "pcm",
        "solvent",
        "smx",
        "scan",
        "van_der_waals",
        "plots",
        "nbo",
        "geom_opt",
        "svp",
        "pcm_nonels",
    ]:
        attr_val = lower_and_check_unique(getattr(qchem, attr))
        setattr(qchem, attr, attr_val)


def get_rem_swaps(rem: dict[str, Any], restart: bool = False) -> dict[str, Any]:
    """
    Automatic swaps for the rem dictionary.

    Parameters
    ----------
    rem
        rem dictionary
    restart
        Whether or not this is a restart calculation.

    Returns
    -------
    dict
        rem dictionary with swaps
    """
    if restart and "scf_guess" not in rem:
        logger.info("Copilot: Setting scf_guess in `rem` to 'read'")
        rem["scf_guess"] = "read"
    if "max_scf_cycles" not in rem:
        rem["max_scf_cycles"] = 200 if rem.get("scf_algorithm") == "gdm" else 100
        logger.info(
            f"Copilot: Setting max_scf_cycles in `rem` to {rem['max_scf_cycles']}"
        )

    return rem



"""Base jobs for ORCA"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ase.calculators.orca import ORCA, OrcaProfile, OrcaTemplate

from quacc import SETTINGS
from quacc.atoms.core import get_final_atoms_from_dyn
from quacc.runners.ase import run_calc, run_opt
from quacc.schemas.ase import summarize_opt_run
from quacc.schemas.cclib import cclib_summarize_run
from quacc.utils.dicts import recursive_dict_merge
from quacc.utils.lists import merge_list_params

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any

    from ase.atoms import Atoms

    from quacc.schemas._aliases.cclib import cclibASEOptSchema, cclibSchema

_LABEL = OrcaTemplate()._label  # skipcq: PYL-W0212
LOG_FILE = f"{_LABEL}.out"
GEOM_FILE = f"{_LABEL}.xyz"


def base_fn(
    atoms: Atoms,
    charge: int = 0,
    spin_multiplicity: int = 1,
    default_inputs: list[str] | None = None,
    default_blocks: list[str] | None = None,
    input_swaps: list[str] | None = None,
    block_swaps: list[str] | None = None,
    additional_fields: dict[str, Any] | None = None,
    copy_files: str | Path | list[str | Path] | None = None,
) -> cclibSchema:
    """
    Base job function for ORCA recipes.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    default_inputs
        Default input parameters.
    default_blocks
        Default block input parameters.
    input_swaps
        List of orcasimpleinput swaps for the calculator. To remove entries
        from the defaults, put a `#` in front of the name.
    block_swaps
        List of orcablock swaps for the calculator. To remove entries
        from the defaults, put a `#` in front of the name.
    additional_fields
        Any additional fields to supply to the summarizer.
    copy_files
        File(s) to copy to the runtime directory. If a directory is provided, it will be recursively unpacked.

    Returns
    -------
    cclibSchema
        Dictionary of results from [quacc.schemas.cclib.cclib_summarize_run][]
    """
    inputs = merge_list_params(default_inputs, input_swaps)
    blocks = merge_list_params(default_blocks, block_swaps)
    orcasimpleinput = " ".join(inputs)
    orcablocks = " ".join(blocks)

    atoms.calc = ORCA(
        profile=OrcaProfile(SETTINGS.ORCA_CMD),
        charge=charge,
        mult=spin_multiplicity,
        orcasimpleinput=orcasimpleinput,
        orcablocks=orcablocks,
    )
    atoms = run_calc(atoms, geom_file=GEOM_FILE, copy_files=copy_files)

    return cclib_summarize_run(atoms, LOG_FILE, additional_fields=additional_fields)


def base_opt_fn(
    atoms: Atoms,
    charge: int = 0,
    spin_multiplicity: int = 1,
    default_inputs: list[str] | None = None,
    default_blocks: list[str] | None = None,
    input_swaps: list[str] | None = None,
    block_swaps: list[str] | None = None,
    opt_defaults: dict[str, Any] | None = None,
    opt_params: dict[str, Any] | None = None,
    additional_fields: dict[str, Any] | None = None,
    copy_files: str | Path | list[str | Path] | None = None,
) -> cclibASEOptSchema:
    """
    Base job function for ORCA recipes with ASE optimizer.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    default_inputs
        Default input parameters.
    default_blocks
        Default block input parameters.
    input_swaps
        List of orcasimpleinput swaps for the calculator. To remove entries
        from the defaults, put a `#` in front of the name.
    block_swaps
        List of orcablock swaps for the calculator. To remove entries
        from the defaults, put a `#` in front of the name.
    opt_defaults
        Default arguments for the ASE optimizer.
    opt_params
        Dictionary of custom kwargs for [quacc.runners.ase.run_opt][]
    additional_fields
        Any additional fields to supply to the summarizer.
    copy_files
        File(s) to copy to the runtime directory. If a directory is provided, it will be recursively unpacked.

    Returns
    -------
    cclibSchema
        Dictionary of results from [quacc.schemas.cclib.cclib_summarize_run][]
    """
    inputs = merge_list_params(default_inputs, input_swaps)
    blocks = merge_list_params(default_blocks, block_swaps)
    opt_flags = recursive_dict_merge(opt_defaults, opt_params)
    orcasimpleinput = " ".join(inputs)
    orcablocks = " ".join(blocks)

    atoms.calc = ORCA(
        profile=OrcaProfile(SETTINGS.ORCA_CMD),
        charge=charge,
        mult=spin_multiplicity,
        orcasimpleinput=orcasimpleinput,
        orcablocks=orcablocks,
    )
    dyn = run_opt(atoms, copy_files=copy_files, **opt_flags)

    final_atoms = get_final_atoms_from_dyn(dyn)
    opt_run_summary = summarize_opt_run(
        dyn,
        charge_and_multiplicity=(charge, spin_multiplicity),
        additional_fields=additional_fields,
    )
    cclib_summary = cclib_summarize_run(
        final_atoms, LOG_FILE, additional_fields=additional_fields
    )
    return recursive_dict_merge(cclib_summary, opt_run_summary)



"""Transition state recipes for the Q-Chem."""
from __future__ import annotations

from typing import TYPE_CHECKING

import psutil
from monty.dev import requires

from quacc import SETTINGS, job, strip_decorator
from quacc.recipes._qchem_legacy._base import base_opt_fn
from quacc.recipes._qchem_legacy.core import relax_job
from quacc.utils.dicts import recursive_dict_merge

try:
    from sella import IRC, Sella

    has_sella = True
except ImportError:
    has_sella = False

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any, Literal

    from ase.atoms import Atoms

    from quacc.schemas._aliases.ase import OptSchema


@job
@requires(has_sella, "Sella must be installed. Refer to the quacc documentation.")
def ts_job(
    atoms: Atoms,
    charge: int,
    spin_multiplicity: int,
    method: str = "wb97mv",
    basis: str = "def2-svpd",
    scf_algorithm: str = "diis",
    pcm_dielectric: str | None = None,
    smd_solvent: str | None = None,
    n_cores: int | None = None,
    overwrite_inputs: dict[str, Any] | None = None,
    opt_params: dict[str, Any] | None = None,
    copy_files: str | Path | list[str | Path] | None = None,
) -> OptSchema:
    """
    TS optimize a molecular structure.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    method
        DFT exchange-correlation functional or other electronic structure
        method. Defaults to wB97M-V.
    basis
        Basis set. Defaults to def2-SVPD.
    scf_algorithm
        Algorithm used to converge the SCF. Defaults to "diis", but for
        particularly difficult cases, "gdm" should be employed instead.
    pcm_dielectric
        Dielectric constant of the optional polarizable continuum implicit
        solvation model. Defaults to None, in which case PCM will not be
        employed.
    smd_solvent
        Solvent to use for SMD implicit solvation model. Examples include
        "water", "ethanol", "methanol", and "acetonitrile". Refer to the Q-Chem
        manual for a complete list of solvents available. Defaults to None, in
        which case SMD will not be employed.
    n_cores
        Number of cores to use for the Q-Chem calculation. Defaults to use all
        cores available on a given node.
    overwrite_inputs
        Dictionary passed to `pymatgen.io.qchem.QChemDictSet` which can modify
        default values set therein as well as set additional Q-Chem parameters.
        See QChemDictSet documentation for more details.
    opt_params
        Dictionary of custom kwargs for [quacc.runners.ase.run_opt][]
    copy_files
        File(s) to copy to the runtime directory. If a directory is provided, it will be recursively unpacked.

    Returns
    -------
    OptSchema
        Dictionary of results from [quacc.schemas.ase.summarize_opt_run][]
    """

    calc_defaults = {
        "basis_set": basis,
        "scf_algorithm": scf_algorithm,
        "method": method,
        "charge": charge,
        "spin_multiplicity": spin_multiplicity,
        "cores": n_cores or psutil.cpu_count(logical=False),
        "qchem_input_params": {
            "pcm_dielectric": pcm_dielectric,
            "smd_solvent": smd_solvent,
            "overwrite_inputs": overwrite_inputs,
            "max_scf_cycles": 200 if scf_algorithm.lower() == "gdm" else 100,
        },
    }
    opt_defaults = {
        "fmax": 0.01,
        "max_steps": 1000,
        "optimizer": Sella,
        "optimizer_kwargs": {"order": 1, "use_TRICs": False},
    }

    if opt_params and opt_params.get("optimizer", Sella) is not Sella:
        raise ValueError("Only Sella should be used for TS optimization.")

    return base_opt_fn(
        atoms,
        charge,
        spin_multiplicity,
        calc_defaults=calc_defaults,
        opt_defaults=opt_defaults,
        opt_params=opt_params,
        additional_fields={"name": "Q-Chem TS"},
        copy_files=copy_files,
    )


@job
@requires(has_sella, "Sella must be installed. Refer to the quacc documentation.")
def irc_job(
    atoms: Atoms,
    charge: int,
    spin_multiplicity: int,
    direction: Literal["forward", "reverse"] = "forward",
    method: str = "wb97mv",
    basis: str = "def2-svpd",
    scf_algorithm: str = "diis",
    pcm_dielectric: str | None = None,
    smd_solvent: str | None = None,
    n_cores: int | None = None,
    overwrite_inputs: dict[str, Any] | None = None,
    opt_params: dict[str, Any] | None = None,
    copy_files: str | Path | list[str | Path] | None = None,
) -> OptSchema:
    """
    IRC optimize a molecular structure.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    direction
        Direction of the IRC. Should be "forward" or "reverse".
    method
        DFT exchange-correlation functional or other electronic structure
        method. Defaults to wB97M-V.
    basis
        Basis set. Defaults to def2-SVPD.
    scf_algorithm
        Algorithm used to converge the SCF. Defaults to "diis", but for
        particularly difficult cases, "gdm" should be employed instead.
    pcm_dielectric
        Dielectric constant of the optional polarizable continuum implicit
        solvation model. Defaults to None, in which case PCM will not be
        employed.
    smd_solvent
        Solvent to use for SMD implicit solvation model. Examples include
        "water", "ethanol", "methanol", and "acetonitrile". Refer to the Q-Chem
        manual for a complete list of solvents available. Defaults to None, in
        which case SMD will not be employed.
    n_cores
        Number of cores to use for the Q-Chem calculation. Defaults to use all
        cores available on a given node.
    overwrite_inputs
        Dictionary passed to `pymatgen.io.qchem.QChemDictSet` which can modify
        default values set therein as well as set additional Q-Chem parameters.
        See QChemDictSet documentation for more details.
    opt_params
        Dictionary of custom kwargs for [quacc.runners.ase.run_opt][]
    copy_files
        File(s) to copy to the runtime directory. If a directory is provided, it will be recursively unpacked.

    Returns
    -------
    OptSchema
        Dictionary of results from [quacc.schemas.ase.summarize_opt_run][]
    """

    calc_defaults = {
        "basis_set": basis,
        "scf_algorithm": scf_algorithm,
        "method": method,
        "charge": charge,
        "spin_multiplicity": spin_multiplicity,
        "cores": n_cores or psutil.cpu_count(logical=False),
        "qchem_input_params": {
            "pcm_dielectric": pcm_dielectric,
            "smd_solvent": smd_solvent,
            "overwrite_inputs": overwrite_inputs,
            "max_scf_cycles": 200 if scf_algorithm.lower() == "gdm" else 100,
        },
    }
    opt_defaults = {
        "fmax": 0.01,
        "max_steps": 1000,
        "optimizer": IRC,
        "optimizer_kwargs": {"keep_going": True},
        "run_kwargs": {"direction": direction},
    }
    if opt_params and opt_params.get("optimizer", IRC) is not IRC:
        raise ValueError("Only Sella's IRC should be used for IRC optimization.")

    return base_opt_fn(
        atoms,
        charge,
        spin_multiplicity,
        calc_defaults=calc_defaults,
        opt_defaults=opt_defaults,
        opt_params=opt_params,
        additional_fields={"name": "Q-Chem IRC"},
        copy_files=copy_files,
    )


@job
@requires(has_sella, "Sella must be installed. Refer to the quacc documentation.")
def quasi_irc_job(
    atoms: Atoms,
    charge: int,
    spin_multiplicity: int,
    direction: Literal["forward", "reverse"] = "forward",
    method: str = "wb97mv",
    basis: str = "def2-svpd",
    scf_algorithm: str = "diis",
    pcm_dielectric: str | None = None,
    smd_solvent: str | None = None,
    n_cores: int | None = None,
    overwrite_inputs: dict[str, Any] | None = None,
    irc_opt_params: dict[str, Any] | None = None,
    relax_opt_params: dict[str, Any] | None = None,
    copy_files: str | Path | list[str | Path] | None = None,
) -> OptSchema:
    """
    Quasi-IRC optimize a molecular structure. Runs `irc_job` for 10 steps (default)
    followed by `relax_job`.

    Parameters
    ----------
    atoms
        Atoms object.
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    direction
        Direction of the IRC. Should be "forward" or "reverse".
    irc_opt_params
        Dictionary of opt_params kwargs for the irc_job.
    relax_opt_params
        Dictionary of opt_params kwargs for the relax_job.
    copy_files
        File(s) to copy to the runtime directory. If a directory is provided, it will be recursively unpacked.

    Returns
    -------
    OptSchema
        Dictionary of results from [quacc.schemas.ase.summarize_opt_run][]
    """

    default_settings = SETTINGS.model_copy()

    irc_opt_params_defaults = {"fmax": 100, "max_steps": 10}
    irc_opt_params = recursive_dict_merge(irc_opt_params_defaults, irc_opt_params)

    SETTINGS.CHECK_CONVERGENCE = False
    irc_summary = strip_decorator(irc_job)(
        atoms,
        charge,
        spin_multiplicity,
        direction=direction,
        method=method,
        basis=basis,
        scf_algorithm=scf_algorithm,
        pcm_dielectric=pcm_dielectric,
        smd_solvent=smd_solvent,
        n_cores=n_cores,
        overwrite_inputs=overwrite_inputs,
        opt_params=irc_opt_params,
        copy_files=copy_files,
    )

    SETTINGS.CHECK_CONVERGENCE = default_settings.CHECK_CONVERGENCE
    relax_summary = strip_decorator(relax_job)(
        irc_summary["atoms"],
        charge,
        spin_multiplicity,
        method=method,
        basis=basis,
        scf_algorithm=scf_algorithm,
        pcm_dielectric=pcm_dielectric,
        smd_solvent=smd_solvent,
        n_cores=n_cores,
        overwrite_inputs=overwrite_inputs,
        opt_params=relax_opt_params,
    )

    relax_summary["initial_irc"] = irc_summary

    return relax_summary



"""Base jobs for DFTB+"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ase.calculators.dftb import Dftb

from quacc import SETTINGS
from quacc.runners.ase import run_calc
from quacc.schemas.ase import summarize_run
from quacc.utils.dicts import recursive_dict_merge
from quacc.utils.files import check_logfile

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any

    from ase.atoms import Atoms

    from quacc.schemas._aliases.ase import RunSchema

LOG_FILE = "dftb.out"
GEOM_FILE = "geo_end.gen"


def base_fn(
    atoms: Atoms,
    calc_defaults: dict[str, Any] | None = None,
    calc_swaps: dict[str, Any] | None = None,
    additional_fields: dict[str, Any] | None = None,
    copy_files: str | Path | list[str | Path] | None = None,
) -> RunSchema:
    """
    Base job function for DFTB+ recipes.

    Parameters
    ----------
    atoms
        Atoms object
    calc_defaults
        The default calculator parameters to use.
    calc_swaps
        Dictionary of custom kwargs for the calculator that would override the
        calculator defaults. Set a value to `quacc.Remove` to remove a pre-existing key
        entirely. For a list of available keys, refer to the
        `ase.calculators.dftb.Dftb` calculator.
    additional_fields
        Any additional fields to supply to the summarizer.
    copy_files
        File(s) to copy to the runtime directory. If a directory is provided, it will be recursively unpacked.

    Returns
    -------
    RunSchema
        Dictionary of results, specified in [quacc.schemas.ase.summarize_run][]
    """

    calc_flags = recursive_dict_merge(calc_defaults, calc_swaps)

    atoms.calc = Dftb(**calc_flags)
    final_atoms = run_calc(atoms, geom_file=GEOM_FILE, copy_files=copy_files)

    if SETTINGS.CHECK_CONVERGENCE:
        if check_logfile(LOG_FILE, "SCC is NOT converged"):
            msg = f"SCC is not converged in {LOG_FILE}"
            raise RuntimeError(msg)
        if calc_flags.get("Driver_") == "GeometryOptimization" and not check_logfile(
            LOG_FILE, "Geometry converged"
        ):
            msg = f"Geometry optimization did not complete in {LOG_FILE}"
            raise RuntimeError(msg)

    return summarize_run(
        final_atoms, input_atoms=atoms, additional_fields=additional_fields
    )



"""Atoms handling with Phonopy"""
from __future__ import annotations

from typing import TYPE_CHECKING

from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.io.phonopy import get_phonopy_structure, get_pmg_structure

try:
    import phonopy
except ImportError:
    phonopy = None

if TYPE_CHECKING:
    from ase.atoms import Atoms
    from numpy.typing import ArrayLike
    from phonopy import Phonopy
    from phonopy.structure.atoms import PhonopyAtoms


def atoms_to_phonopy(
    atoms: Atoms,
    supercell_matrix: ArrayLike,
    atom_disp: float,
    phonopy_kwargs: dict | None = None,
) -> Phonopy:
    """
    Convert an ASE atoms object to a phonopy object with displacements
    generated.

    Parameters
    ----------
    atoms
        ASE atoms object
    supercell_matrix
        Supercell matrix to use.
    atom_disp
        Atomic displacement (A).
    phonopy_kwargs
        Additional kwargs to pass to the Phonopy class.

    Returns
    -------
    Phonopy
        Phonopy object
    """
    phonopy_kwargs = phonopy_kwargs or {}
    structure = AseAtomsAdaptor().get_structure(atoms)
    phonopy_atoms = get_phonopy_structure(structure)
    phonon = phonopy.Phonopy(
        phonopy_atoms, supercell_matrix=supercell_matrix, **phonopy_kwargs
    )
    phonon.generate_displacements(distance=atom_disp)
    return phonon


def phonopy_atoms_to_ase_atoms(phonpy_atoms: PhonopyAtoms) -> Atoms:
    """
    Convert a phonopy atoms object to an ASE atoms object.

    Parameters
    ----------
    phonpy_atoms
        Phonopy atoms object

    Returns
    -------
    Atoms
        ASE atoms object
    """
    pmg_structure = get_pmg_structure(phonpy_atoms)
    return AseAtomsAdaptor().get_atoms(pmg_structure)



"""Schemas for VASP."""
from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import TYPE_CHECKING

from emmet.core.tasks import TaskDoc
from pymatgen.command_line.bader_caller import bader_analysis_from_path
from pymatgen.command_line.chargemol_caller import ChargemolAnalysis

from quacc import SETTINGS
from quacc.schemas.ase import summarize_run
from quacc.utils.dicts import clean_task_doc
from quacc.wflow_tools.db import results_to_db

if TYPE_CHECKING:
    from typing import Any

    from ase.atoms import Atoms
    from maggma.core import Store
    from pymatgen.core import Structure

    from quacc.schemas._aliases.vasp import BaderSchema, ChargemolSchema, VaspSchema

logger = logging.getLogger(__name__)


def vasp_summarize_run(
    atoms: Atoms,
    dir_path: str | Path | None = None,
    prep_next_run: bool = True,
    run_bader: bool | None = None,
    run_chargemol: bool | None = None,
    check_convergence: bool = True,
    additional_fields: dict[str, Any] | None = None,
    store: Store | None = None,
) -> VaspSchema:
    """
    Get tabulated results from a VASP run and store them in a database-friendly format.

    Parameters
    ----------
    atoms
        ASE Atoms object following a calculation.
    dir_path
        Path to VASP outputs. A value of None specifies the current working
        directory
    prep_next_run
        Whether the Atoms object stored in {"atoms": atoms} should be prepared
        for the next run. This clears out any attached calculator and moves the
        final magmoms to the initial magmoms.
    run_bader
        Whether a Bader analysis should be performed. Will not run if bader
        executable is not in PATH even if bader is set to True. Defaults to
        VASP_BADER in settings.
    run_chargemol
        Whether a Chargemol analysis should be performed. Will not run if chargemol
        executable is not in PATH even if chargmeol is set to True. Defaults to
        VASP_CHARGEMOL in settings.
    check_convergence
        Whether to throw an error if convergence is not reached.
    additional_fields
        Additional fields to add to the task document.
    store
        Maggma Store object to store the results in. If None,
        `SETTINGS.PRIMARY_STORE` will be used.

    Returns
    -------
    VaspSchema
        Dictionary representation of the task document
    """

    additional_fields = additional_fields or {}
    run_bader = SETTINGS.VASP_BADER if run_bader is None else run_bader
    run_chargemol = SETTINGS.VASP_CHARGEMOL if run_chargemol is None else run_chargemol
    dir_path = dir_path or Path.cwd()
    store = SETTINGS.PRIMARY_STORE if store is None else store

    # Fetch all tabulated results from VASP outputs files. Fortunately, emmet
    # already has a handy function for this
    vasp_task_doc = TaskDoc.from_directory(dir_path).model_dump()
    struct = vasp_task_doc["output"]["structure"]

    # Check for calculation convergence
    if check_convergence and vasp_task_doc["state"] != "successful":
        raise RuntimeError(
            f"VASP calculation did not converge. Will not store task data. Refer to {dir_path}"
        )

    base_task_doc = summarize_run(atoms, prep_next_run=prep_next_run, store=False)

    # Get Bader analysis
    if run_bader:
        try:
            bader_results = _bader_runner(dir_path, structure=struct)
        except Exception:
            bader_results = None
            logging.warning("Bader analysis could not be performed.", exc_info=True)

        if bader_results:
            vasp_task_doc["bader"] = bader_results[0]
            struct = bader_results[1]

    # Get the Chargemol analysis
    if run_chargemol:
        try:
            chargemol_results = _chargemol_runner(dir_path, structure=struct)
        except Exception:
            chargemol_results = None
            logging.warning("Chargemol analysis could not be performed.", exc_info=True)

        if chargemol_results:
            vasp_task_doc["chargemol"] = chargemol_results[0]
            struct = chargemol_results[1]

    # Override the Structure to have the attached properties
    vasp_task_doc["output"]["structure"] = struct

    # Make task document
    unsorted_task_doc = base_task_doc | vasp_task_doc | additional_fields
    task_doc = clean_task_doc(unsorted_task_doc)

    # Store the results
    if store:
        results_to_db(store, task_doc)

    return task_doc


def _bader_runner(
    path: str | None = None, structure: Structure = None
) -> tuple[BaderSchema, Structure | None]:
    """
    Runs a Bader partial charge and spin moment analysis using the VASP output
    files in the given path. This function requires that `bader` is located in
    your PATH environment variable. See
    http://theory.cm.utexas.edu/henkelman/code/bader for the bader code.

    Parameters
    ----------
    path
        The path where the VASP output files are located. Must include CHGCAR,
        AECCAR0, AECCAR2, and POTCAR files. These files can be gzip'd or not --
        it doesn't matter. If None, the current working directory is used.
    structure
        The structure object to attach the Bader charges and spins to.

    Returns
    -------
    BaderSchema
        Dictionary containing the Bader analysis summary
    Structure
        Structure object with the Bader charges and spins attached
    """
    path = path or Path.cwd()

    # Make sure files are present
    relevant_files = ["AECCAR0", "AECCAR2", "CHGCAR", "POTCAR"]
    for f in relevant_files:
        if not Path(path, f).exists() and not Path(path, f"{f}.gz").exists():
            msg = f"Could not find {f} in {path}."
            raise FileNotFoundError(msg)

    bader_stats = bader_analysis_from_path(path)

    # Store the partial charge, which is much more useful than the raw charge
    # and is more intuitive than the charge transferred. An atom with a positive
    # partial charge is cationic, whereas an atom with a negative partial charge
    # is anionic.
    bader_stats["partial_charges"] = [-c for c in bader_stats["charge_transfer"]]

    # Some cleanup of the returned dictionary
    if "magmom" in bader_stats:
        bader_stats["spin_moments"] = bader_stats["magmom"]
    for k in ["charge", "charge_transfer", "reference_used", "magmom"]:
        bader_stats.pop(k, None)

    # Attach the Bader charges and spins to the structure
    if structure:
        structure.add_site_property("bader_charge", bader_stats["partial_charges"])
        if "spin_moments" in bader_stats:
            structure.add_site_property("bader_spin", bader_stats["spin_moments"])

    return bader_stats, structure


def _chargemol_runner(
    path: str | None = None,
    atomic_densities_path: str | None = None,
    structure: Structure | None = None,
) -> tuple[ChargemolSchema, Structure | None]:
    """
    Runs a Chargemol (i.e. DDEC6 + CM5) analysis using the VASP output files in
    the given path. This function requires that the chargemol executable, given
    by the name `Chargemol_09_26_2017_linux_parallel`,
    `Chargemol_09_26_2017_linux_serial`, or `chargemol` is in the system PATH
    environment variable. See https://sourceforge.net/projects/ddec/files for
    the Chargemol code.

    Parameters
    ----------
    path
        The path where the VASP output files are located. Must include CHGCAR,
        AECCAR0, AECCAR2, and POTCAR files. These files can be gzip'd or not --
        it doesn't matter. If None, the current working directory is used.
    atomic_densities_path
        The path where the reference atomic densities are located for Chargemol.
        If None, we assume that this directory is defined in an environment
        variable named DDEC6_ATOMIC_DENSITIES_DIR. See the Chargemol
        documentation for more information.

    Returns
    -------
    ChargemolSchema
        Dictionary containing the Chargemol analysis summary
    Structure
        Structure object with the Chargemol charges and spins attached
    """
    path = path or Path.cwd()

    # Make sure files are present
    relevant_files = ["AECCAR0", "AECCAR2", "CHGCAR", "POTCAR"]
    for f in relevant_files:
        if not Path(path, f).exists() and not Path(path, f"{f}.gz").exists():
            msg = f"Could not find {f} in {path}."
            raise FileNotFoundError(msg)

    # Check environment variable
    if atomic_densities_path is None and "DDEC6_ATOMIC_DENSITIES_DIR" not in os.environ:
        msg = "DDEC6_ATOMIC_DENSITIES_DIR environment variable not defined."
        raise OSError(msg)

    # Run Chargemol analysis
    chargemol_stats = ChargemolAnalysis(
        path=path, atomic_densities_path=atomic_densities_path
    )

    # Attach the Chargemol charges and spins to the structure
    if structure:
        structure.add_site_property(
            "ddec6_charge", chargemol_stats["ddec"]["partial_charges"]
        )
        if "spin_moments" in chargemol_stats["ddec"]:
            structure.add_site_property(
                "ddec6_spin", chargemol_stats["ddec"]["spin_moments"]
            )
        if "cm5" in chargemol_stats:
            structure.add_site_property(
                "cm5_charge", chargemol_stats["cm5"]["partial_charges"]
            )

    return chargemol_stats, structure



"""Core recipes for espresso."""
from __future__ import annotations

from typing import TYPE_CHECKING

from quacc import job
from quacc.calculators.espresso.espresso import EspressoTemplate
from quacc.recipes.espresso._base import base_fn

if TYPE_CHECKING:
    from pathlib import Path

    from ase.atoms import Atoms

    from quacc.schemas._aliases.ase import RunSchema


@job
def static_job(
    atoms: Atoms,
    preset: str | None = "sssp_1.3.0_pbe_efficiency",
    parallel_info: dict[str] | None = None,
    copy_files: str | Path | list[str | Path] | None = None,
    **calc_kwargs,
) -> RunSchema:
    """
    Function to carry out a basic SCF calculation with pw.x.

    Parameters
    ----------
    atoms
        The Atoms object.
    preset
        The name of a YAML file containing a list of parameters to use as
        a "preset" for the calculator. quacc will automatically look in the
        `ESPRESSO_PRESET_DIR` (default: quacc/calculators/espresso/presets).
    parallel_info
        Dictionary containing information about the parallelization of the
        calculation. See the ASE documentation for more information.
    copy_files
        List of files to copy to the calculation directory. Useful for copying
        files from a previous calculation. This parameter can either be a string
        or a list of strings.

        If a string is provided, it is assumed to be a path to a directory,
        all of the child tree structure of that directory is going to be copied to the
        scratch of this calculation. For phonon_job this is what most users will want to do.

        If a list of strings is provided, each string point to a specific file. In this case
        it is important to note that no directory structure is going to be copied, everything
        is copied at the root of the temporary directory.
    **calc_kwargs
        Additional keyword arguments to pass to the Espresso calculator. See the
        docstring of [quacc.calculators.espresso.espresso.Espresso][] for more information.

    Returns
    -------
    RunSchema
        Dictionary of results from [quacc.schemas.ase.summarize_run][]
    """

    calc_defaults = {"input_data": {"control": {"calculation": "scf"}}}

    return base_fn(
        atoms,
        preset=preset,
        template=EspressoTemplate("pw"),
        calc_defaults=calc_defaults,
        calc_swaps=calc_kwargs,
        parallel_info=parallel_info,
        additional_fields={"name": "pw.x Static"},
        copy_files=copy_files,
    )


@job
def relax_job(
    atoms: Atoms,
    preset: str | None = "sssp_1.3.0_pbe_efficiency",
    relax_cell: bool = False,
    parallel_info: dict[str] | None = None,
    copy_files: str | Path | list[str | Path] | None = None,
    **calc_kwargs,
) -> RunSchema:
    """
    Function to carry out a structure relaxation with pw.x.

    Parameters
    ----------
    atoms
        The Atoms object.
    preset
        The name of a YAML file containing a list of parameters to use as
        a "preset" for the calculator. quacc will automatically look in the
        `ESPRESSO_PRESET_DIR` (default: quacc/calculators/espresso/presets).
    relax_cell
        Whether to relax the cell or not.
    parallel_info
        Dictionary containing information about the parallelization of the
        calculation. See the ASE documentation for more information.
    copy_files
        List of files to copy to the calculation directory. Useful for copying
        files from a previous calculation. This parameter can either be a string
        or a list of strings.

        If a string is provided, it is assumed to be a path to a directory,
        all of the child tree structure of that directory is going to be copied to the
        scratch of this calculation. For phonon_job this is what most users will want to do.

        If a list of strings is provided, each string point to a specific file. In this case
        it is important to note that no directory structure is going to be copied, everything
        is copied at the root of the temporary directory.
    **calc_kwargs
        Additional keyword arguments to pass to the Espresso calculator. See the
        docstring of [quacc.calculators.espresso.espresso.Espresso][] for more information.

    Returns
    -------
    RunSchema
        Dictionary of results from [quacc.schemas.ase.summarize_run][]
    """

    calc_defaults = {
        "input_data": {
            "control": {"calculation": "vc-relax" if relax_cell else "relax"}
        }
    }

    return base_fn(
        atoms,
        preset=preset,
        template=EspressoTemplate("pw"),
        calc_defaults=calc_defaults,
        calc_swaps=calc_kwargs,
        parallel_info=parallel_info,
        additional_fields={"name": "pw.x Relax"},
        copy_files=copy_files,
    )


@job
def post_processing_job(
    prev_dir: str | Path, parallel_info: dict[str] | None = None, **calc_kwargs
) -> RunSchema:
    """
    Function to carry out a basic pp.x calculation (post-processing).
    It is mainly used to extract the charge density from a previous pw.x calculation.
    and perform simple to complex post-processing on it. Fore more details please see
    (https://www.quantum-espresso.org/Doc/INPUT_PP.html)

    Parameters
    ----------
    prev_dir
        Outdir of the previously ran pw.x calculation. This is used to copy
        the entire tree structure of that directory to the working directory
        of this calculation.
    parallel_info
        Dictionary containing information about the parallelization of the
        calculation. See the ASE documentation for more information.
    **calc_kwargs
        calc_kwargs dictionary possibly containing the following keys:

        - input_data: dict
        - additional_fields: list[str] | str

        See the docstring of ase.io.espresso.write_fortran_namelist for more information.

    Returns
    -------
    RunSchema
        Dictionary of results from [quacc.schemas.ase.summarize_run][]
    """

    calc_defaults = {
        "input_data": {
            "inputpp": {"plot_num": 0},
            "plot": {
                "iflag": 3,
                "output_format": 6,
                "fileout": "pseudo_charge_density.cube",
            },
        }
    }

    return base_fn(
        template=EspressoTemplate("pp"),
        calc_defaults=calc_defaults,
        calc_swaps=calc_kwargs,
        parallel_info=parallel_info,
        additional_fields={"name": "pp.x post-processing"},
        copy_files=prev_dir,
    )
